/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2015 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.editor;

import static org.fortiss.tooling.base.layout.DefaultLayoutConstants.DEFAULT_GRID_SIZE;
import static org.fortiss.tooling.base.utils.LayoutDataUtils.moveNode;

import org.eclipse.draw2d.FigureCanvas;
import org.eclipse.draw2d.geometry.Point;
import org.eclipse.draw2d.geometry.Rectangle;
import org.eclipse.gef.GraphicalViewer;
import org.eclipse.gef.KeyHandler;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyEvent;
import org.fortiss.tooling.base.model.layout.ILayoutedModelElement;
import org.fortiss.tooling.base.ui.editpart.PositionedEditPartBase;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * A {@link KeyHandler} for Diagram editors.
 * 
 * Provides arrow keys handling:
 * - if an element is selected, then using the arrows will move this element
 * - if no element is selected, then using the arrows will scroll the diagram
 * In both cases, pressing the Shift key at the same time allows to move/scroll faster.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: CC77BCDDC8E3E2E4D8C7C94535532E4B
 */
public class DiagramKeyHandler extends KeyHandler {

	/** The embedded viewer. */
	private GraphicalViewer viewer;

	/** Constructor. */
	public DiagramKeyHandler(GraphicalViewer viewer) {
		this.viewer = viewer;
	}

	/** {@inheritDoc} */
	@Override
	public boolean keyPressed(KeyEvent event) {
		if(event.keyCode != SWT.ARROW_DOWN && event.keyCode != SWT.ARROW_UP &&
				event.keyCode != SWT.ARROW_LEFT && event.keyCode != SWT.ARROW_RIGHT) {
			return false;
		}
		if(scrollViewerIfNeeded(event)) {
			return true;
		}
		return scrollElementIfNeeded(event);
	}

	/** Scrolls the viewer according to the event if no editpart is selected. */
	private boolean scrollViewerIfNeeded(KeyEvent event) {
		if(!(viewer.getSelectedEditParts().isEmpty() && viewer.getControl() instanceof FigureCanvas))
			return false;
		FigureCanvas figCanvas = (FigureCanvas)viewer.getControl();
		Point loc = figCanvas.getViewport().getViewLocation();
		Rectangle area = figCanvas.getViewport().getClientArea(Rectangle.SINGLETON).scale(.1);
		Point scrollCoeff = computeScrollCoeff(event);
		figCanvas.scrollToY(loc.y + scrollCoeff.y * area.height);
		figCanvas.scrollToX(loc.x + scrollCoeff.x * area.width);
		return true;
	}

	/** Scrolls the model of the selected editpart according to the event. */
	private boolean scrollElementIfNeeded(KeyEvent event) {
		if(!(viewer.getFocusEditPart() instanceof PositionedEditPartBase<?>))
			return false;
		final ILayoutedModelElement model =
				(ILayoutedModelElement)viewer.getFocusEditPart().getModel();
		ITopLevelElement topLevel = IPersistencyService.getInstance().getTopLevelElementFor(model);
		final Point scrollCoeff = computeScrollCoeff(event);
		scrollCoeff.scale(DEFAULT_GRID_SIZE);
		topLevel.runAsCommand(new Runnable() {
			@Override
			public void run() {
				moveNode(model, scrollCoeff.x, scrollCoeff.y);
			}
		});
		return true;
	}

	/** Returns a coefficient of 1 if the SHIFT key is not pressed, 50 otherwise. */
	private Point computeScrollCoeff(KeyEvent event) {
		int coeff = (event.stateMask & SWT.SHIFT) == 0 ? 1 : 10;
		if(event.keyCode == SWT.ARROW_DOWN)
			return new Point(0, coeff);
		if(event.keyCode == SWT.ARROW_UP)
			return new Point(0, -coeff);
		coeff *= (viewer.getControl().getStyle() & SWT.MIRRORED) != 0 ? -1 : 1;
		if(event.keyCode == SWT.ARROW_LEFT)
			return new Point(-coeff, 0);
		if(event.keyCode == SWT.ARROW_RIGHT)
			return new Point(coeff, 0);
		return null;
	}
}
