/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.dnd;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.base.model.element.IConnector;
import org.fortiss.tooling.base.model.layout.Point;
import org.fortiss.tooling.kernel.extension.data.IConnectionCompositionContext;

/**
 * A {@link ConnectionDragContext} stores information obtained from user's drag
 * gesture, i.e. the source edit part and the location w.r.t. this edit part and
 * the target edit part and the location w.r.t. this edit part.
 * <p>
 * This information is usually provided to the connection compositor in order to provide the UI data
 * necessary to construct layout information based on the drag gesture positions.
 * <p>
 * Note that this context is used for drag operations within a single editor, e.g. from one
 * displayed edit part to another in the same diagram.
 * 
 * @see ConnectionDropContext
 * @see ElementDropContext
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: 98762A374BEB623A786D0CE1ED5059AB
 */
public class ConnectionDragContext implements IConnectionCompositionContext {

	/** Source {@link EObject}. */
	private EObject sourceElement;

	/** Parent {@link EObject} of the source {@link EObject}. */
	private EObject sourceParent;

	/** Source position. */
	private Point sourcePosition;

	/** Predicate whether the source element is a root object. */
	private boolean isSourceRoot;

	/** Target {@link EObject}. */
	private EObject targetElement;

	/** Parent {@link EObject} of the target {@link EObject}. */
	private EObject targetParent;

	/** Target position. */
	private Point targetPosition;

	/** Predicate whether the target element is a root object. */
	private boolean isTargetRoot;

	/** Predicate whether the source of the connection is a free, unconnected {@link IConnector}. */
	private boolean isSourceFreeConnector;

	/** Predicate whether the target of the connection is a free, unconnected {@link IConnector}. */
	private boolean isTargetFreeConnector;

	/** Indicator whether the intended connection is "valid". */
	private boolean isValidSource;

	/** Caches the zoom factor of the context. */
	private double zoom;

	/** Constructor. */
	public ConnectionDragContext(EObject source, EObject sourceParent, Point sourcePosition,
			EObject target, EObject targetParent, Point targetPosition,
			boolean isSourceFreeConnector, boolean isSourceRoot, boolean isTargetFreeConnector,
			boolean isValidSource, boolean isTargetRoot) {
		this.sourceElement = source;
		this.sourceParent = sourceParent;
		this.sourcePosition = sourcePosition;
		this.targetElement = target;
		this.targetParent = targetParent;
		this.targetPosition = targetPosition;
		this.isSourceFreeConnector = isSourceFreeConnector;
		this.isSourceRoot = isSourceRoot;
		this.isTargetFreeConnector = isTargetFreeConnector;
		this.isValidSource = isValidSource;
		this.isTargetRoot = isTargetRoot;
	}

	/** Constructor. */
	public ConnectionDragContext(EObject source, EObject parent, Point sourcePosition,
			boolean isSourceRoot, boolean isSourceFreeConnector, boolean isValidSource) {
		this(source, parent, sourcePosition, null, null, null, isSourceFreeConnector, isSourceRoot,
				false, isValidSource, false);
	}

	/**
	 * Set target data.
	 */
	public void setTarget(EObject target, EObject parent, Point targetPosition,
			boolean isTargetFreeConnector, boolean isTargetRoot) {
		this.targetElement = target;
		this.targetParent = parent;
		this.targetPosition = targetPosition;
		this.isTargetFreeConnector = isTargetFreeConnector;
		this.isTargetRoot = isTargetRoot;
	}

	/** Returns the source position. */
	public Point getSourcePosition() {
		return sourcePosition;
	}

	/** Returns the source {@link EObject}. */
	public EObject getSource() {
		return sourceElement;
	}

	/** Returns the parent {@link EObject} of the source {@link EObject}. */
	public EObject getSourceParent() {
		return sourceParent;
	}

	/** Returns the target position. */
	public Point getTargetPosition() {
		return targetPosition;
	}

	/** Returns the target {@link EObject}. */
	public EObject getTarget() {
		return targetElement;
	}

	/** Returns the parent {@link EObject} of the target {@link EObject}. */
	public EObject getTargetParent() {
		return targetParent;
	}

	/** Returns the zoom factor that was used when creating the context. */
	public double getZoom() {
		return zoom;
	}

	/** Returns whether drop target is root edit part (base). */
	public boolean isTargetRoot() {
		return isTargetRoot;
	}

	/** Returns whether drop source is root edit part (base). */
	public boolean isSourceRoot() {
		return isSourceRoot;
	}

	/** Returns whether the source of the connection is a free, unconnected {@link IConnector}. */
	public boolean isSourceFreeConnector() {
		return isSourceFreeConnector;
	}

	/** Returns whether the target of the connection is a free, unconnected {@link IConnector}. */
	public boolean isTargetFreeConnector() {
		return isTargetFreeConnector;
	}

	/** Indicates whether the intended connection is "valid". */
	// FIXME (FH): todo without issue number
	// TODO: This method replicates the "old" behavior. Evaluate whether it is not advisable to
	// make the validity checks on creation an eliminate this method.
	public boolean isValidSource() {
		return isValidSource;
	}

	/** Returns the corresponding source {@link ElementDropContext}. */
	public ElementDropContext getSourceDropContext() {
		return new ElementDropContext(getSource(), getSourcePosition(), isSourceRoot, zoom);
	}

	/** Returns the corresponding source {@link ElementDropContext}. */
	public ElementDropContext getTargetDropContext() {
		return new ElementDropContext(getTarget(), getTargetPosition(), isTargetRoot, zoom);
	}
}
