/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2015 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.ui.extension.base;

import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.jobs.Job;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.resource.ImageDescriptor;
import org.eclipse.swt.SWT;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Group;
import org.eclipse.swt.widgets.Shell;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.ErrorVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.FailVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedVerificationStatus;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService.IFix;
import org.fortiss.tooling.kernel.ui.extension.IConstraintVerifierUI;
import org.fortiss.tooling.kernel.ui.service.IConstraintVerificationUIService;
import org.fortiss.tooling.kernel.ui.service.IModelElementHandlerService;
import org.fortiss.tooling.kernel.ui.util.ConstraintsUIUtils;
import org.fortiss.tooling.kernel.utils.ConstraintsUtils;

/**
 * Base classes for constraint verification GUI.
 * 
 * @author vincent
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: A75818A660AD2BE3AF8BDB6B56E4E818
 */
public class ConstraintVerifierUIBases {

	/** Base class for constraint verification GUI. */
	public static abstract class ConstraintVerifierUIBase implements IConstraintVerifierUI {
		/** {@inheritDoc} */
		@Override
		public boolean canOpen(ConstraintInstance constraint, IConstraintVerificationStatus status) {
			return status instanceof OutdatedVerificationStatus;
		}

		/** {@inheritDoc} */
		@Override
		public boolean openStatus(ConstraintInstance c, IConstraintVerificationStatus status) {
			if(!canOpen(c, status)) {
				return false;
			}
			if(status instanceof OutdatedVerificationStatus) {
				IConstraintVerificationService.getInstance().verify(c);
				IConstraintVerificationStatus currentStatus =
						IConstraintVerificationUIService.getInstance().getStatus(c);
				// FIXME: bit-wise AND surely not intended
				if(currentStatus instanceof ErrorVerificationStatus & canOpen(c, status)) {
					openStatus(c, status);
				}
				return true;
			}
			if(status instanceof FailVerificationStatus ||
					status instanceof ErrorVerificationStatus) {
				List<IFix> fixes = IConstraintVerificationService.getInstance().fixes(c);
				Shell shell = Display.getCurrent().getActiveShell();
				int result = new FixDialog(shell, status, getMessage(c, status), fixes).open();
				if(fixes.size() == 1 && result == 0) {
					// get(0) because the size is equal to 1 by the previous test
					fixes.get(0).runFix(status);
				} else if(result > 1) {
					// TODO: comment 'result - 2'
					fixes.get(result - 2).runFix(status);
				}
				return true;
			}
			return false;
		}

		/** {@inheritDoc} */
		@Override
		public boolean isTimeConsuming() {
			return false;
		}

		/** Dialog presenting the error/failure and the possible fix(es). */
		private static class FixDialog extends MessageDialog {
			/** List of possible fixes. */
			List<IFix> fixes;

			/** Constructor. */
			public FixDialog(Shell parentShell, IConstraintVerificationStatus s, String msg,
					List<IFix> fixes) {
				super(parentShell, getTitle(s), null, msg, getIconType(s), getButtons(fixes), 1);
				this.fixes = fixes;
			}

			// TODO: remove undocumented params
			/**
			 * @param fixes
			 * @return Buttons to display: "OK" is always displayed. In addition, if there is *only
			 *         one fix, we provide one button for this fix. If there is more than one
			 *         possible fixes, they will displayed instead horizontally (via getCustomArea).
			 */
			private static String[] getButtons(List<IFix> fixes) {
				String[] res = new String[fixes.size() == 1 ? 2 : 1];
				if(fixes.size() == 1) {
					// get(0) because the size is equal to 1 by the previous test
					res[0] = "Fix: " + fixes.get(0).getDescription();
				}
				res[res.length - 1] = "OK";
				return res;
			}

			/**
			 * @param s
			 * @return the icon type corresponding to <code>s</code>.
			 */
			private static String getTitle(IConstraintVerificationStatus s) {
				if(s instanceof FailVerificationStatus) {
					return "Verification of constraint failed";
				}
				return "Error during constraint verification";
			}

			/**
			 * @param s
			 * @return the icon type corresponding to <code>s</code>.
			 */
			private static int getIconType(IConstraintVerificationStatus s) {
				return s instanceof FailVerificationStatus ? MessageDialog.WARNING : ERROR;
			}

			/** {@inheritDoc} */
			@Override
			protected Control createCustomArea(Composite parent) {
				if(fixes.size() <= 1) {
					return null;
				}
				Group buttonGroup = new Group(parent, SWT.NONE);
				buttonGroup.setLayout(new GridLayout());
				buttonGroup.setLayoutData(new GridData(SWT.RIGHT, SWT.TOP, true, false));
				buttonGroup.setText("Possible fixes");
				for(IFix fix : fixes) {
					Button button = new Button(buttonGroup, SWT.None);
					button.setText(fix.getDescription());
					button.addSelectionListener(new SelectionAdapter() {
						@Override
						public void widgetSelected(SelectionEvent event) {
							buttonPressed(fixes.indexOf(fix) + 1);
						}
					});
				}
				// Does not matter what we return as long as it's not null
				return buttonGroup;
			}
		}

		/** {@inheritDoc} */
		@Override
		public String getMessage(ConstraintInstance c, IConstraintVerificationStatus status) {
			IConstraintVerificationStatus s =
					IConstraintVerificationUIService.getInstance().getStatus(c);
			// Default message. Inheriting classes SHOULD specialize.
			String name = IModelElementHandlerService.getInstance().getName(s.getConstraint());
			String pre =
					s instanceof ErrorVerificationStatus ? "Error while checking" : "Unsatisfied";
			return pre + " constraint" + (name == null ? "." : ": \"" + name + "\"");
		}

		/** {@inheritDoc} */
		@Override
		public void onOutdate(ConstraintInstance constraint) {
			// By default, we just set the status to outdated
			// TODO: import static
			ConstraintsUtils.createOutdatedVerificationStatus(constraint);
			ConstraintsUIUtils.triggerMarkersRefresh(constraint);
		}

		/** {@inheritDoc} */
		@Override
		public boolean shouldBeManuallyActivated() {
			return false;
		}

		/** {@inheritDoc} */
		@Override
		public boolean displayAsWarning() {
			// By default, constraints failures are displayed as error not as warnings
			return false;
		}

		/** {@inheritDoc} */
		@Override
		public ImageDescriptor getIconImageDescriptor() {
			return null;
		}
	}

	/**
	 * Base class for constraint verifier UIs which automatically trigger a new check when the
	 * constraint gets outdated.
	 */
	public static abstract class ConstraintVerifierUIBaseAutocheck extends ConstraintVerifierUIBase {

		/** {@inheritDoc} */
		@Override
		public void onOutdate(ConstraintInstance constraint) {
			Display display = Display.getDefault();
			String name = "Verifying " + ConstraintVerifierUIBaseAutocheck.this.getDescription();
			Job latestJob = new Job(name) {
				/** {@inheritDoc} */
				@Override
				protected IStatus run(IProgressMonitor monitor) {
					try {
						display.asyncExec(() -> IConstraintVerificationService.getInstance()
								.cancel(constraint));
						IConstraintVerificationService.getInstance().verify(constraint);
						// TODO: import static
						display.asyncExec(() -> ConstraintsUIUtils
								.triggerMarkersRefresh(constraint));
						return Status.OK_STATUS;
					} catch(IllegalArgumentException e) {
						// This sort of exception might happen, not sure why, but we *in no case*
						// want to throw an error message to the user in case of some failure
						// resulting of a constraint check which he did not explicitly triggered
						return Status.CANCEL_STATUS;
					}
				}
			};
			// If the constraint is considered time consuming, we wait 1s before starting the job
			// because user modifications can go faster than the check.
			latestJob.schedule(isTimeConsuming() ? 1000 : 0);
		}

		/** {@inheritDoc} */
		@Override
		public boolean canOpen(ConstraintInstance constraint, IConstraintVerificationStatus status) {
			return false;
		}
	}

	/** Handler for constraints. */
	public static class ConstraintHandler extends ModelElementHandlerBase<ConstraintInstance> {

		/** {@inheritDoc} */
		@Override
		public String getName(ConstraintInstance element) {
			return getDescription(element);
		}

		/** {@inheritDoc} */
		@Override
		public ImageDescriptor getIconImageDescriptor() {
			// This method will actually never be called
			return null;
		}

		/** {@inheritDoc} */
		@Override
		public ImageDescriptor getIconImageDescriptor(ConstraintInstance element) {
			String id = element.getConstraintID();
			return IConstraintVerificationUIService.getInstance().getIconImageDescriptor(id);
		}
	}
}
