package org.fortiss.tooling.kernel.ui.internal;

import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getBundle;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.getConfigurationElements;
import static org.fortiss.tooling.kernel.utils.ExtensionPointUtils.loadClass;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.warning;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.function.Predicate;
import java.util.stream.Collectors;

import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.emf.common.notify.Adapter;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.ecore.ENamedElement;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.EReference;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.jface.resource.ImageDescriptor;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.model.constraints.ConstrainedWithChecksum;
import org.fortiss.tooling.kernel.model.constraints.ConstraintInstance;
import org.fortiss.tooling.kernel.model.constraints.ConstraintsPackage;
import org.fortiss.tooling.kernel.model.constraints.IConstrained;
import org.fortiss.tooling.kernel.model.constraints.IConstraintInstanceContainer;
import org.fortiss.tooling.kernel.model.constraints.IConstraintVerificationStatus;
import org.fortiss.tooling.kernel.model.constraints.OutdatedVerificationStatus;
import org.fortiss.tooling.kernel.service.IConstraintVerificationService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.ui.ToolingKernelUIActivator;
import org.fortiss.tooling.kernel.ui.extension.IConstraintVerifierUI;
import org.fortiss.tooling.kernel.ui.internal.introspection.items.ConstraintVerificationUIServiceIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.ui.service.IConstraintVerificationUIService;
import org.fortiss.tooling.kernel.ui.util.ConstraintsUIUtils;
import org.fortiss.tooling.kernel.utils.EcoreUtils;
import org.fortiss.tooling.kernel.utils.LoggingUtils;
import org.osgi.framework.Bundle;

/**
 * Implementation of {@link IConstraintVerificationService}.
 * 
 * @author aravantinos
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: C98962AD9F3FF617BA1EF29E5FC20960
 */
public final class ConstraintVerificationUIService implements IConstraintVerificationUIService,
		IIntrospectiveKernelService {

	/** The singleton service instance. */
	private static ConstraintVerificationUIService INSTANCE = new ConstraintVerificationUIService();

	/** Stores the class to handler map. */
	// TODO: use Class<? extends IConstraitnVerifierUI> as first dimension
	// (see TutorialUIService.instanceMap
	protected final Map<String, IConstraintVerifierUI> handlerMap =
			new HashMap<String, IConstraintVerifierUI>();

	/** Returns the service instance. */
	public static ConstraintVerificationUIService getInstance() {
		return INSTANCE;
	}

	/** Activates logging to observe when constraints get outdated. */
	private static final boolean DEBUG = false;

	/** Starts the service. */
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
		List<ITopLevelElement> tops = IPersistencyService.getInstance().getTopLevelElements();
		for(ITopLevelElement top : tops) {
			TreeIterator<EObject> tree = top.getRootModelElement().eAllContents();
			while(tree.hasNext()) {
				EObject next = tree.next();
				if(next instanceof ConstraintInstance) {
					install((ConstraintInstance)next);
				}
			}
		}
	}

	/** Registers the verifier UI part with the service. */
	@Override
	public void registerConstraintVerifierUI(IConstraintVerifierUI verifier) {
		addHandler(verifier.getID(), verifier);
	}

	/** Initializes the service by setting up the handler map. */
	public void initializeService() {
		setupHandlerMap();
	}

	/** Initializes the handler map from plugin extensions. */
	private void setupHandlerMap() {
		for(IConfigurationElement ce : getConfigurationElements(getExtensionPointName(),
				getConfigurationElementName())) {
			Bundle bundle = getBundle(ce);
			try {
				Class<?> handlerClass =
						loadClass(ce.getAttribute(getHandlerClassAttribute()), bundle);
				IConstraintVerifierUI handler =
						(IConstraintVerifierUI)handlerClass.getConstructor().newInstance();
				addHandler(handler.getID(), handler);
			} catch(Exception ex) {
				error(ToolingKernelActivator.getDefault(), ex.getMessage(), ex);
			}
		}
	}

	/** Adds the given handler to the map and the set. */
	protected void addHandler(String id, IConstraintVerifierUI handler) {
		IConstraintVerifierUI existingVerifier = handlerMap.get(id);
		if(existingVerifier != null) {
			warning(ToolingKernelActivator.getDefault(),
					"Encountered more than one handler registered with " + getExtensionPointName() +
							"(" + handler.getClass() + "," + existingVerifier.getClass() + ")");
		}
		handlerMap.put(id, handler);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service provides the UI complement of the \"pure\" constraint verification service." +
				"\n\nIt provides the following services:" +
				"\n- inform whether a constraint instance can be open or not," +
				"\n- open the status of a constraint instance," +
				"\n- \"install\" a constraint instance, i.e., sets listeners on the constrained element(s) in order" +
				"\n  to automatically detect when a constraint is outdated," +
				"\n- inform, for a given constraint instance, whether it should display an error or a warning in case of non-satisfaction," +
				"\n- provide a user-friendly short description for a constraint (useful in various automatic messages)," +
				"\n- provide an icon for a given constraint (useful in the context menu)," +
				"\n- give the status of a given constraint instance - taking into account whether the constraint is activated or not," +
				"\n- activate a given constraint," +
				"\n- deactivate a given constraint," +
				"\n- give the list of all available constraints," +
				"\n- give the list of constraints which should always be activated - by opposition to those which should be explicitly activated by the user," +
				"\n- give the list of constraints which are currently activated - by the user or automatically," +
				"\n- complete a given project with instances of a given constraint if necessary." +
				"\n\nThis service is \"fed\" by constraint verifier UI's." +
				"\n\nThe class org.fortiss.tooling.kernel.ui.extension.base.ConstraintVerifierUIBaseAutocheck should" +
				"\ngenerally provide a sufficient initial implementation. You are however strongly encouraged" +
				"\nto extend this class to provide a better feedback to the user (e.g., with detailed errors)." +
				"\n\nSee the developer wiki page \"Add a new constraint on a metamodel\"" +
				" for detailed documentation on developing a new sort of constraint." +
				"\n\nThe service extension point is '" + getExtensionPointName() + "'.";
	}

	/**
	 * @param constraint
	 * @return the first registered verifier which provides a GUI for the verifier of
	 *         <code>constraint</code>
	 */
	private IConstraintVerifierUI getFirstVerifier(ConstraintInstance constraint) {
		return getFirstVerifier(constraint.getConstraintID());
	}

	/**
	 * @param id
	 * @return the first registered verifier which can verify a constraint of type <code>id</code>.
	 */
	protected IConstraintVerifierUI getFirstVerifier(String id) {
		IConstraintVerifierUI existingVerifier = handlerMap.get(id);
		if(existingVerifier == null) {
			LoggingUtils.error(ToolingKernelActivator.getDefault(),
					String.format("Cannot find verifier UI for constraint %s", id));
			return null;
		}
		return existingVerifier;
	}

	/** {@inheritDoc} */
	@Override
	public void openStatus(ConstraintInstance c) {
		IConstraintVerifierUI verifier = getFirstVerifier(c);
		verifier.openStatus(c, getStatus(c));
		ConstraintsUIUtils.triggerMarkersRefresh(c);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canOpen(ConstraintInstance c) {
		IConstraintVerifierUI verifier = getFirstVerifier(c);
		return verifier.canOpen(c, getStatus(c));
	}

	/** {@inheritDoc} */
	@Override
	public void install(ConstraintInstance constraint) {
		// TODO: import static
		ConstraintsUIUtils.triggerMarkersRefresh(constraint);
		// TODO: remember a reference for cleanup?
		new ConstrainedAdapters(constraint);
	}

	/** Adapter for changes on constrained elements. */
	private class ConstrainedAdapters {

		/** Map containing the adapter of each constrained element. */
		Map<IConstrained, Adapter> constrainedAdapterMap;

		/** Constructor. */
		public ConstrainedAdapters(ConstraintInstance constraint) {
			constrainedAdapterMap = new HashMap<IConstrained, Adapter>();
			for(ConstrainedWithChecksum cwc : constraint.getConstrainedsWithChecksum()) {
				IConstrained constrained = cwc.getConstrained();
				Adapter adapter = new EContentAdapter() {
					@Override
					public void notifyChanged(Notification notification) {
						super.notifyChanged(notification);
						if(getStatus(constraint) == null || isTouch(notification)) {
							return;
						}
						if(constraintRemoved(notification, constrained)) {
							return;
						}
						constraintOutdated(constraint, notification);
					}
				};
				constrained.eAdapters().add(adapter);
				constrainedAdapterMap.put(constrained, adapter);
			}
		}

		/**
		 * @param constraint
		 * @param notification
		 *            Checks if <code>notification</code> denotes a change entailing that
		 *            <code>constraint</code> is outdated. If so updates the status to outdated or
		 *            triggers an automatic check if the <code>autocheck</code> field of the
		 *            constraint is <code>true</code>.
		 * @return <code>true</code> iff <code>notification</code> indeed entailed that
		 *         <code>constraint</code> is outdated.
		 */
		private boolean
				constraintOutdated(ConstraintInstance constraint, Notification notification) {
			if(skipStatusChangeNotification(constraint, notification)) {
				return false;
			}
			if(!IConstraintVerificationService.getInstance().isUpToDate(constraint)) {
				if(DEBUG) {
					String msg = "Constraint " + constraint.toString() + " is outdated.";
					LoggingUtils.info(ToolingKernelUIActivator.getDefault(), msg);
				}
				getFirstVerifier(constraint).onOutdate(constraint);
				return true;
			}
			return false;
		}

		/**
		 * @param constraint
		 * @param notification
		 * @return true iff <code>notification</code> denotes a change of the
		 *         <code>constraint</code>'s status.
		 */
		private boolean skipStatusChangeNotification(ConstraintInstance constraint,
				Notification notification) {
			String n =
					ConstraintsPackage.eINSTANCE.getConstraintInstance_VerificationStatus()
							.getName();
			if(notification.getNotifier() != null && notification.getNotifier().equals(constraint)) {
				if(notification.getFeature() instanceof EReference) {
					return(n.equals(((ENamedElement)notification.getFeature()).getName()));
				}
			}
			return false;
		}

		/**
		 * @param notification
		 * @param constrained
		 * @return true if the notification is a consequence of a constraint of
		 *         <code>constrained</code> having been removed.
		 */
		private boolean constraintRemoved(Notification notification, IConstrained constrained) {
			if(notification.getEventType() != Notification.REMOVE) {
				return false;
			}
			if(!notification.getNotifier().equals(constrained)) {
				return false;
			}
			if(notification.getFeature() instanceof EReference) {
				String name = ((ENamedElement)notification.getFeature()).getName();
				if(name.equals(ConstraintsPackage.eINSTANCE.getIConstrained_Constraints().getName())) {
					constrained.eAdapters().remove(constrainedAdapterMap.get(constrained));
					constrainedAdapterMap.remove(constrained);
					return true;
				}
			}
			return true;
		}
	}

	/**
	 * @param notification
	 * @return true if <code>notification</code> is not relevant and should be skipped
	 */
	private static boolean isTouch(Notification notification) {
		// Notifications which do not notify about a model change.
		if(notification.isTouch()) {
			return true;
		}
		// Notifications initiated by the the AF3 kernel which are only GUI related.
		if(notification.getEventType() == EcoreUtils.KERNEL_EMF_EVENT_TYPE_COUNT) {
			return true;
		}
		return false;
	}

	/** Returns the extension point name. */
	protected String getExtensionPointName() {
		return "org.fortiss.tooling.kernel.ui.constraintVerifierUI";
	}

	/** Returns the configuration element name. */
	protected String getConfigurationElementName() {
		return "constraintVerifierUI";
	}

	/** Returns the handler class attribute name. */
	protected String getHandlerClassAttribute() {
		return "constraintVerifierUI";
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new ConstraintVerificationUIServiceIntrospectionDetailsItem(handlerMap);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Constraint Verification UI Service";
	}

	/** Activated constraints per top level element. */
	private static Map<ITopLevelElement, Set<String>> activatedConstraints = new HashMap<>();

	/** {@inheritDoc} */
	@Override
	public IConstraintVerificationStatus getStatus(ConstraintInstance c) {
		ITopLevelElement top = IPersistencyService.getInstance().getTopLevelElementFor(c);
		boolean isActivated = false;
		Set<String> set = activatedConstraints.get(top);
		if(set != null) {
			for(String id : set) {
				if(id.equals(c.getConstraintID())) {
					isActivated = true;
				}
			}
		}
		return isActivated ? c.getVerificationStatus() : null;
	}

	/** {@inheritDoc} */
	@Override
	public Set<String> getAllConstraints() {
		return handlerMap.keySet();
	}

	/** {@inheritDoc} */
	@Override
	public Set<String> getActivatedConstraints(ITopLevelElement top) {
		Set<String> res = activatedConstraints.get(top);
		return res != null ? res : new HashSet<>();
	}

	/** {@inheritDoc} */
	@Override
	public void activate(String id, IConstraintInstanceContainer cstrContainer) {
		ITopLevelElement top =
				IPersistencyService.getInstance().getTopLevelElementFor(cstrContainer);
		if(!activatedConstraints.containsKey(top)) {
			activatedConstraints.put(top, new HashSet<>());
		}
		activatedConstraints.get(top).add(id);
		List<ConstraintInstance> addeds =
				IConstraintVerificationService.getInstance()
						.addMissingConstraintsInstances(top, id);
		if(!addeds.isEmpty() && getFirstVerifier(id).isTimeConsuming()) {
			// TODO: import static
			ConstraintsUIUtils.displayTimeConsumptionWarning(id);
		}
		// Install notifiers for newly added constraints
		for(ConstraintInstance added : addeds) {
			IConstraintVerificationUIService.getInstance().install(added);
		}
		EList<ConstraintInstance> cstrs = cstrContainer.getConstraintInstances();
		Predicate<ConstraintInstance> isID = c -> id.equals(c.getConstraintID());
		List<ConstraintInstance> idCstrs = cstrs.stream().filter(isID).collect(Collectors.toList());
		update(idCstrs);
	}

	/** {@inheritDoc} */
	@Override
	public void addMissingConstraintInstances(IConstraintInstanceContainer cstrContainer) {
		ITopLevelElement top =
				IPersistencyService.getInstance().getTopLevelElementFor(cstrContainer);
		List<ConstraintInstance> addeds = new ArrayList<ConstraintInstance>();
		for(String id : getActivatedConstraints(top)) {
			addeds.addAll(IConstraintVerificationService.getInstance()
					.addMissingConstraintsInstances(top, id));
			IConstraintVerificationService.getInstance().addMissingConstraintsInstances(top, id);
		}
		// Install notifiers for newly added constraints
		for(ConstraintInstance added : addeds) {
			IConstraintVerificationUIService.getInstance().install(added);
		}
		update(addeds);
	}

	/** Checks the constraints of type <code>id</code> among the given list of constraints. */
	private void update(List<ConstraintInstance> constraints) {
		// Checks all the constraints: the newly added and the already existing ones which are now
		// activated.
		for(ConstraintInstance c : constraints) {
			// No need to check if it turns out the status is already known from a previous
			// activation.
			if(!IConstraintVerificationService.getInstance().isUpToDate(c) ||
					c.getVerificationStatus() instanceof OutdatedVerificationStatus) {
				getFirstVerifier(c).onOutdate(c);
			}
		}
		ConstraintsUIUtils.triggerMarkerRefresh(constraints);
	}

	/** {@inheritDoc} */
	@Override
	public void deactivate(String id, IConstraintInstanceContainer cstrContainer) {
		ITopLevelElement top =
				IPersistencyService.getInstance().getTopLevelElementFor(cstrContainer);
		activatedConstraints.get(top).remove(id);
		ConstraintsUIUtils.triggerMarkerRefresh(id, cstrContainer);
	}

	/** {@inheritDoc} */
	@Override
	public String getDescription(String id) {
		IConstraintVerifierUI verifier = getFirstVerifier(id);
		return verifier == null ? null : verifier.getDescription();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public ImageDescriptor getIconImageDescriptor(String id) {
		// TODO: use ConstraintInstance and call getConstraintID() internally
		IConstraintVerifierUI verifier = getFirstVerifier(id);
		return verifier == null ? null : verifier.getIconImageDescriptor();
	}

	/** {@inheritDoc} */
	@Override
	public boolean shallDisplayAsWarning(ConstraintInstance constraint) {
		return getFirstVerifier(constraint).displayAsWarning();
	}

	/** {@inheritDoc} */
	@Override
	public Set<String> getAlwaysActivatedConstraints() {
		HashSet<String> res = new HashSet<>();
		for(Entry<String, IConstraintVerifierUI> entry : handlerMap.entrySet()) {
			if(!entry.getValue().shouldBeManuallyActivated()) {
				res.add(entry.getKey());
			}
		}
		return res;
	}
}
