/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2011 fortiss GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal.storage.eclipse;

import static org.eclipse.emf.ecore.util.EcoreUtil.resolveAll;
import static org.fortiss.tooling.kernel.utils.EMFResourceUtils.buildOptionsMap;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.warning;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.lang.reflect.Method;
import java.util.EventObject;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.common.command.AbstractCommand;
import org.eclipse.emf.common.command.BasicCommandStack;
import org.eclipse.emf.common.command.CommandStack;
import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.util.BasicEList;
import org.eclipse.emf.common.util.Diagnostic;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.util.Diagnostician;
import org.eclipse.emf.ecore.util.EContentAdapter;
import org.eclipse.emf.ecore.xmi.XMIResource;
import org.eclipse.emf.ecore.xml.type.AnyType;
import org.eclipse.emf.transaction.TransactionalEditingDomain;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.model.FortissToolingKernelPackage;
import org.fortiss.tooling.kernel.model.IIdLabeled;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.utils.UniqueIDUtils;

/**
 * The model context provides additional commands and hooks for a model.
 * Additionally an editing domain is provided.
 * 
 * @author hummel
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: D2DD5730F29EF8C125811F8C1B4B8D7F
 */
class ModelContext implements ITopLevelElement, CommandStackListener {

	/** The resource set we are using. */
	private final ResourceSet rset;

	/** The file this model is from. */
	private final IFile file;

	/** The resource containing the model. */
	private final Resource resource;

	/** The editing domain used for this model. */
	private final TransactionalEditingDomain editingDomain;

	/** The transactional command stack. */
	private final AutoUndoCommandStack transactionalCommandStack;

	/** Stores the command stack listener for this context. */
	private final List<CommandStackListener> listeners = new LinkedList<CommandStackListener>();

	/** Flag for remembering whether the last change of the file was intended. */
	private boolean lastChangeWasIntended = false;

	/** The maximal ID used in this model (used to generate new IDs). */
	private int maxId = -1;

	/**
	 * There might be mismatches between the meta-model of a file and the meta-model of AF3
	 * (typically when using a file saved with an older version of AF3). In this case, the parser
	 * will parse the file (or fail if the file is really too different from what is expected), but
	 * it will also return the XML nodes ("features") that it failed to parse. These are stored in a
	 * map "EObject to AnyType": EObject gives the context in which the unknown feature has been met
	 * and AnyType provides the actual unknown data. If non-empty, the migrators have to deal with
	 * it. If they don't the model context is considered buggy.
	 */
	private Map<EObject, AnyType> unknownFeatures;

	/** Constructor. */
	/* package */ModelContext(IFile file) throws IOException {
		this.file = file;

		editingDomain = TransactionalEditingDomain.Factory.INSTANCE.createEditingDomain();
		rset = editingDomain.getResourceSet();

		URI uri = URI.createPlatformResourceURI(file.getFullPath().toString(), true);
		resource = getResourceSet().createResource(uri);

		resource.load(buildOptionsMap());

		if(resource instanceof XMIResource) {
			unknownFeatures = ((XMIResource)resource).getEObjectToExtensionMap();
		} else {
			unknownFeatures = new HashMap<EObject, AnyType>();
		}

		transactionalCommandStack = new AutoUndoCommandStack(editingDomain);
		transactionalCommandStack.addCommandStackListener(this);

		// Install adapter to maintain XMI IDs
		resource.eAdapters().add(new EContentAdapter() {
			@Override
			public void notifyChanged(Notification notification) {
				super.notifyChanged(notification);

				int eventType = notification.getEventType();
				Object feature = notification.getFeature();

				boolean featureOk = feature == FortissToolingKernelPackage.Literals.IID_LABELED__ID;
				if(eventType == Notification.SET && featureOk) {

					// Update XMI IDs when Kernel IDs are changed by calls of IdLabled.setId()
					Object notifier = notification.getNotifier();

					if(resource instanceof XMIResource && notifier instanceof EObject) {
						String newVal = notification.getNewStringValue();
						((XMIResource)resource).setID((EObject)notifier, newVal);
					}
				}
			}
		});

		checkIDs();
	}

	/** {@inheritDoc} */
	@Override
	public String getId(EObject modelElement) {
		return resource instanceof XMIResource ? ((XMIResource)resource).getID(modelElement) : null;
	}

	/** {@inheritDoc} */
	@Override
	public EObject getRootModelElement() {
		// get(0) cause the root model element is always the first one
		return resource.getContents().get(0);
	}

	/** {@inheritDoc} */
	@Override
	public void addCommandStackListener(CommandStackListener listener) {
		if(!listeners.contains(listener)) {
			listeners.add(listener);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeCommandStackListener(CommandStackListener listener) {
		listeners.remove(listener);
	}

	/**
	 * Checks whether all IDs are present and unique and updates {@link #maxId}. In case duplicate
	 * IDs have been detected (which could only be caused by a bug in AF3), these are removed and an
	 * error is logged.
	 */
	private void checkIDs() {
		Set<Integer> ids = new HashSet<Integer>();

		boolean hadMissing = false;
		boolean hadDuplicates = false;
		for(Iterator<EObject> i = getRootModelElement().eAllContents(); i.hasNext();) {
			EObject eo = i.next();
			if(eo instanceof IIdLabeled) {
				final IIdLabeled element = (IIdLabeled)eo;
				int id = element.getId();
				if(id <= 0) {
					hadMissing = true;
				} else {
					// Reset duplicate IDs. They will be regenerated together with missing IDs. This
					// is a safe operation since at this stage, references are still based on Java
					// references. Only in the persisted resource, references will be based on IDs,
					// which is why special care must be taken to ensure uniqueness.
					if(ids.contains(id)) {
						hadDuplicates = true;
						runAsCommand(new Runnable() {
							@Override
							public void run() {
								element.setId(0);
							}
						});
					} else {
						ids.add(id);
					}
				}
				maxId = Math.max(maxId, id);
			}
		}

		maxId = Math.max(0, maxId);

		if(hadMissing || hadDuplicates) {
			runAsCommand(new Runnable() {
				@Override
				public void run() {
					maxId = UniqueIDUtils.generateMissingIDs(getRootModelElement(), maxId);
				}
			});
		}

		if(hadDuplicates) {
			String msg = "Duplicate IDs have been removed from \"";
			msg += resource.getURI().lastSegment() + "\". ";
			msg += "Please report this incident since it could result in corrupted model files.";
			error(ToolingKernelActivator.getDefault(), msg);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void prepareIDs(EObject other) {
		maxId = UniqueIDUtils.prepareIDs(other, getRootModelElement(), maxId);
	}

	/** Destroys this context and closes all dependent editors. */
	public void destroy() {
		// discard changes
		((BasicCommandStack)editingDomain.getCommandStack()).saveIsDone();
		transactionalCommandStack.removeCommandStackListener(this);
	}

	/** Returns the file associated with this model. */
	public IFile getFile() {
		return file;
	}

	/** Returns whether the model is dirty (i.e. if there are unsaved changes). */
	@Override
	public boolean isDirty() {
		if(lastChangeWasIntended) {
			return false;
		}
		return ((BasicCommandStack)editingDomain.getCommandStack()).isSaveNeeded();
	}

	/** Perform saving of the model. */
	@Override
	public synchronized void doSave(IProgressMonitor monitor) throws IOException, CoreException {
		monitor.beginTask("Saving...", rset.getResources().size() + 2);

		// Step 1: Validate resource managed by kernel
		// Perform ID checking before safe to not produce inconsistent models
		checkIDs();

		// Perform diagnostics
		Diagnostic validate = Diagnostician.INSTANCE.validate(getRootModelElement());
		if(validate.getSeverity() != Diagnostic.OK) {
			warning(ToolingKernelActivator.getDefault(), getFullMessage(validate));
		}
		monitor.worked(1);

		// We use a two-phase save process, as otherwise inconsistencies in the model can lead to
		// partially written files which are unreadable and thus cause data loss.

		// Step 2: first we write the content of the resource to a byte array this will abort the
		// save process if something is wrong with the EMF model (a typical example of that is an
		// EObject that is referenced from somewhere, but is not aggregated in the model tree (its
		// eContainer() is null).

		// Preserve exceptions
		Exception exception = null;

		ByteArrayOutputStream bytes = new ByteArrayOutputStream();
		Map<String, Object> saveOptions = buildOptionsMap();
		try {
			resource.save(bytes, saveOptions);
		} catch(IOException e) {
			exception = e;
		} finally {
			// avoid reloading due to disk content change
			setLastChangeWasCausedByModelContext();
			monitor.worked(1);

			try {
				// Step 3: write file content to disk (for resource managed by kernel)
				ByteArrayInputStream stream = new ByteArrayInputStream(bytes.toByteArray());
				file.setContents(stream, false, true, monitor);
			} catch(CoreException e) {
				if(exception == null) {
					exception = e;
				}
				monitor.worked(1);
			}
		}

		// Step 3..n: Save resources not managed by kernel
		for(Resource currentResource : rset.getResources()) {
			if(currentResource != resource && !editingDomain.isReadOnly(currentResource)) {
				try {
					currentResource.save(saveOptions);
				} catch(IOException e) {
					if(exception == null) {
						exception = e;
					}
				} finally {
					monitor.worked(1);
				}
			}
		}
		monitor.done();

		if(exception == null) {
			// Set save state and notify listeners
			((BasicCommandStack)editingDomain.getCommandStack()).saveIsDone();
			performNotifyListeners(editingDomain.getCommandStack());

			IPersistencyService ps = IPersistencyService.getInstance();
			if(ps != null) {
				ps.notifyTopLevelElementChanged(ModelContext.this);
			}
		} else {
			// Throw first caught IOException or CoreException
			if(exception instanceof IOException) {
				throw (IOException)exception;
			} else if(exception instanceof CoreException) {
				throw (CoreException)exception;
			}

			// Other exceptions are not caught
			assert (false);
		}
	}

	/**
	 * Gathers all the error messages of a given diagnostic.
	 * 
	 * @param diagnostic
	 * @return The error message of <code>diagnostic</code> <i>and the ones of its potential
	 *         children</i>.
	 */
	private String getFullMessage(Diagnostic diagnostic) {
		String res = new String();
		if(diagnostic.getSeverity() != Diagnostic.OK) {
			res = diagnostic.getMessage();
			for(Diagnostic child : diagnostic.getChildren()) {
				String childMessage = getFullMessage(child);
				res += (childMessage.isEmpty() ? "" : "\n") + childMessage;
			}
		}
		return res;
	}

	/**
	 * Perform saving of the model and states the last change as intended.
	 * Added to solve issue #2140.
	 */
	@Override
	public void doSaveAndResetLastChangeWasIntended(IProgressMonitor monitor) throws IOException,
			CoreException {
		doSave(monitor);
		getAndResetLastChangeWasCausedByModelContext();
	}

	/**
	 * Returns whether the last change was caused by a save. This method has a
	 * side effect! The {@link #lastChangeWasIntended} flag will be reset, so
	 * this method should only be called once per change event.
	 */
	/* package */synchronized boolean getAndResetLastChangeWasCausedByModelContext() {
		final boolean result = lastChangeWasIntended;
		lastChangeWasIntended = false;
		return result;
	}

	/** Sets the {@link #lastChangeWasIntended} flag to true. */
	/* package */synchronized void setLastChangeWasCausedByModelContext() {
		lastChangeWasIntended = true;
	}

	/** Runs the given runnable as a command. */
	@Override
	public void runAsCommand(final Runnable runnable) {
		transactionalCommandStack.execute(new AbstractCommand() {

			@Override
			public boolean canExecute() {
				return true;
			}

			@Override
			public void execute() {
				runnable.run();
			}

			@Override
			public void redo() {
				// we do nothing here, as redo is handled by our command stack.
			}
		});
	}

	/** Runs the given runnable as a command that does not make the model dirty. */
	@Override
	public void runAsNonDirtyingCommand(final Runnable runnable) {
		try {
			transactionalCommandStack.executeNonDirtyingNonUndoing(new AbstractCommand() {
				@Override
				public void execute() {
					runnable.run();
				}

				@Override
				public boolean canExecute() {
					return true;
				}

				@Override
				public void redo() {
					// we do nothing here, as redo is handled by our command stack.
				}

				/** {@inheritDoc} */
				@Override
				public boolean canUndo() {
					return false;
				}
			});
		} catch(Exception e) {
			e.printStackTrace();
			String msg = "Problem with synchronizing library and model!";
			error(ToolingKernelActivator.getDefault(), msg, e);
		}
	}

	/** Helper method for calling notifyListeners on the given stack. */
	/* package */void performNotifyListeners(CommandStack commandStack) {
		try {
			Method notifyListenersMethod =
					BasicCommandStack.class.getDeclaredMethod("notifyListeners");
			notifyListenersMethod.setAccessible(true);
			notifyListenersMethod.invoke(commandStack);
		} catch(Exception e) {
			error(ToolingKernelActivator.getDefault(), "Notification after save failed!", e);
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean canUndo() {
		return transactionalCommandStack.canUndo();
	}

	/** {@inheritDoc} */
	@Override
	public boolean canRedo() {
		return transactionalCommandStack.canRedo();
	}

	/** {@inheritDoc} */
	@Override
	public void undo() {
		transactionalCommandStack.undo();
	}

	/** {@inheritDoc} */
	@Override
	public void redo() {
		transactionalCommandStack.redo();
	}

	/** {@inheritDoc} */
	@Override
	public String getSaveableName() {
		return getFile().getName();
	}

	/** {@inheritDoc} */
	@Override
	public void commandStackChanged(EventObject event) {
		EventObject relayEvent = new EventObject(this);
		for(CommandStackListener l : listeners) {
			l.commandStackChanged(relayEvent);
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean canDelete() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean delete() {

		// Determine resources referenced by the resource managed by the kernel
		EList<Resource> resourcesToDelete = new BasicEList<Resource>();
		resolveAll(rset);
		for(Resource currentResource : rset.getResources()) {
			if(currentResource != resource) {
				resourcesToDelete.add(currentResource);
			}
		}

		boolean rval = true;

		// First, delete resource managed by the kernel
		try {
			getFile().delete(false, null);
		} catch(CoreException e) {
			String msg = "Error during deletion of model file " + getFile().getName() + ".";
			error(ToolingKernelActivator.getDefault(), msg);
			rval = false;
		}

		// The, delete all other resources
		for(Resource currentResource : resourcesToDelete) {
			try {
				currentResource.delete(null);
			} catch(IOException e) {
				String msg = "Error during deletion of model file ";
				msg += currentResource.getURI().lastSegment() + ".";
				error(ToolingKernelActivator.getDefault(), msg);
				rval = false;
			}
		}

		return rval;
	}

	/** {@inheritDoc} */
	@Override
	public ResourceSet getResourceSet() {
		return rset;
	}

	/** Returns unknownFeatures. */
	public Map<EObject, AnyType> getUnknownFeatures() {
		return unknownFeatures;
	}
}
