/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.ide.commons.ui.xmlmodel;

import java.util.ArrayList;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.conqat.lib.commons.assertion.CCSMPre;
import org.conqat.lib.commons.collections.IdentityHashSet;
import org.conqat.lib.commons.xml.XMLUtils;

/**
 * This model is the base class for root models, which contain the entire DOM.
 * 
 * @author doelle
 * @author hummelb
 * @author $Author: deissenb $
 * @version $Rev: 34252 $
 * @levd.rating GREEN Hash: 60EDC5EED44C8B4FB6EAA78102B613A1
 */
public abstract class XmlRootModelBase {

	/** The DOM document this model is based on. */
	protected Document document;

	/**
	 * The root element. This is stored separately, as for some models the root
	 * element for the XML model is not the top-level XML element.
	 */
	protected Element rootElement;

	/** Flag for storing pending notifications. */
	private boolean hadNotificationDuringCompound = false;

	/** Counter for the number of nested operations. */
	private int nestedCompoundOperations = 0;

	/** Set of listeners */
	private final IdentityHashSet<IModelListener> listeners = new IdentityHashSet<IModelListener>();

	/** Creates a {@link XmlRootModelBase} from a XML document. */
	protected XmlRootModelBase(Document document) {
		insertNewDocument(document);
	}

	/** Add listener. */
	public synchronized void addModelListener(IModelListener listener) {
		listeners.add(listener);
	}

	/** Remove listener. */
	public synchronized void removeModelListener(IModelListener listener) {
		listeners.remove(listener);
	}

	/** Fire model changed event. */
	public synchronized void fireModelChanged() {
		if (nestedCompoundOperations > 0) {
			hadNotificationDuringCompound = true;
			return;
		}

		for (IModelListener listener : new ArrayList<IModelListener>(listeners)) {
			listener.modelChanged();
		}
	}

	/** Returns the document of this model. */
	public Document getDocument() {
		return document;
	}

	/** Returns the root element of this model. */
	public Element getRootElement() {
		return rootElement;
	}

	/**
	 * Inserts a new document to be used by this model.
	 * <p>
	 * This method is also called by the constructor, so certain attributes in
	 * subclasses might not be initialized when this is called!
	 */
	public void insertNewDocument(Document document) {
		this.document = document;
		rootElement = document.getDocumentElement();

		fireModelChanged();
	}

	/**
	 * Begins a composite operation, which will be observed as a single
	 * operation by the user in terms of refresh, undo, etc. Note that composite
	 * operations should be closed by {@link #endCompositeOperation()} in any
	 * case (using try/finally). Nesting of composite operations is possible.
	 */
	public synchronized void beginCompositeOperation() {
		++nestedCompoundOperations;
	}

	/**
	 * Ends a composite operation started by {@link #beginCompositeOperation()}.
	 * Calling this without a previous {@link #beginCompositeOperation()} is an
	 * error (checked by assertion).
	 */
	public synchronized void endCompositeOperation() {
		CCSMPre.isTrue(nestedCompoundOperations > 0,
				"No composite operations open!");
		--nestedCompoundOperations;

		// fire pending notifications
		if (nestedCompoundOperations == 0) {
			if (hadNotificationDuringCompound) {
				hadNotificationDuringCompound = false;
				fireModelChanged();
			}
		}
	}

	/** Create new DOM element in the correct namespace. */
	public Element createElement(String name) {
		return document.createElementNS(document.getDocumentElement()
				.getNamespaceURI(), name);
	}

	/**
	 * Set the text content of the given element's first child that is an
	 * element named as specified. If none exists yet, a new one is created
	 * first. The {@link #fireModelChanged()} method will be called at the end.
	 */
	protected void setNamedChildContent(Element element, String name,
			String content) {
		Element child = XMLUtils.getNamedChild(element, name);
		if (child == null) {
			child = createElement(name);
			element.appendChild(child);
		}
		child.setTextContent(content);
		fireModelChanged();
	}

}