/*-------------------------------------------------------------------------+
|                                                                          |
| Copyright 2005-2011 The ConQAT Project                                   |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+-------------------------------------------------------------------------*/
package org.conqat.lib.commons.date;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.Collection;
import java.util.Collections;
import java.util.Date;

import org.conqat.lib.commons.error.NeverThrownRuntimeException;
import org.conqat.lib.commons.factory.IFactory;

/**
 * Utility methods for working on date objects.
 * 
 * @author juergens
 * @author $Author: deissenb $
 * @version $Rev: 35183 $
 * @ConQAT.Rating GREEN Hash: A11396F7CD6F8CD45E1F4517250DD04B
 */
public class DateUtils {

	/** The factory used to create the date in {@link #getNow()}. */
	private static IFactory<Date, NeverThrownRuntimeException> nowFactory;

	/** Returns the latest date in a collection of dates */
	public static Date getLatest(Collection<Date> dates) {
		if (dates.isEmpty()) {
			return null;
		}
		return Collections.max(dates);
	}

	/** Returns the earliest date in a collection of dates */
	public static Date getEarliest(Collection<Date> dates) {
		if (dates.isEmpty()) {
			return null;
		}
		return Collections.min(dates);
	}

	/** Returns the earlier of two dates, or null, if one of the dates is null */
	public static Date min(Date d1, Date d2) {
		if (d1 == null || d2 == null) {
			return null;
		}

		if (d1.compareTo(d2) < 0) {
			return d1;
		}
		return d2;
	}

	/** Returns the later of two dates or null, if one of the dates is null. */
	public static Date max(Date d1, Date d2) {
		if (d1 == null || d2 == null) {
			return null;
		}

		if (d2.compareTo(d1) > 0) {
			return d2;
		}
		return d1;
	}

	/**
	 * Returns the current time as a {@link Date} object. This is preferred to
	 * directly calling the {@link Date#Date()} constructor, as this method
	 * provides a central entry point that allows the notion of time to be
	 * tweaked (e.g. for testing, when we want a certain date to be returned).
	 * 
	 * The behavior of this method can be affected either by calling the
	 * {@link #testcode_setNowFactory(IFactory)} method, or by providing a fixed
	 * date in format "yyyyMMddHHmmss" as the system property
	 * "org.conqat.lib.commons.date.now".
	 */
	public static synchronized Date getNow() {
		if (nowFactory == null) {
			String property = System
					.getProperty("org.conqat.lib.commons.date.now");
			if (property == null) {
				nowFactory = new CurrentDateFactory();
			} else {
				try {
					nowFactory = new FixedDateFactory(new SimpleDateFormat(
							"yyyyMMddHHmmss").parse(property));
				} catch (ParseException e) {
					// fail hard in case of misconfiguration
					throw new RuntimeException(
							"Invalid date string provided via system property: "
									+ property, e);
				}
			}
		}
		return nowFactory.create();
	}

	/**
	 * Returns the factory that affects the notion of "now" in {@link #getNow()}
	 * . This should be only called from test code!
	 */
	public static synchronized IFactory<Date, NeverThrownRuntimeException> testcode_getNowFactory() {
		return nowFactory;
	}

	/**
	 * Sets the factory that affects the notion of "now" in {@link #getNow()}.
	 * This should be only called from test code!
	 */
	public static synchronized void testcode_setNowFactory(
			IFactory<Date, NeverThrownRuntimeException> nowFactory) {
		DateUtils.nowFactory = nowFactory;
	}

	/** A factory returning a fixed date. */
	public static class FixedDateFactory implements
			IFactory<Date, NeverThrownRuntimeException> {

		/** The date used as now. */
		private final Date now;

		/** Constructor. */
		public FixedDateFactory(Date now) {
			this.now = now;
		}

		/** {@inheritDoc} */
		@Override
		public Date create() throws NeverThrownRuntimeException {
			return (Date) now.clone();
		}
	}

	/** A factory returning the current date. */
	public static class CurrentDateFactory implements
			IFactory<Date, NeverThrownRuntimeException> {

		/** {@inheritDoc} */
		@Override
		public Date create() throws NeverThrownRuntimeException {
			return new Date();
		}
	}
}